import os
from edsl import BASE_DIR


EXAMPLE_RESULTS_PATH = os.path.join(BASE_DIR, "utilities/data/example_results.json")
SCOOTER_RESULTS_PATH = os.path.join(BASE_DIR, "utilities/data/scooter_results.json")
